@@ -98,13 +98,13 @@ class BrailleGenerator(generator.Generator):
             focusedRegion = result[0]
         except:
             focusedRegion = None
+        try:
+            role = obj.getRole()
+        except:
+            role = None
         for region in result:
-            try:
-                role = obj.getRole()
-            except:
-                role = None
             if isinstance(region, (braille.Component, braille.Text)) \
-               and region.accessible == obj:
+               and self._script.utilities.isSameObject(region.accessible, obj, True):
                 focusedRegion = region
                 break
             elif isinstance(region, braille.Text) \
@@ -394,15 +394,11 @@ class BrailleGenerator(generator.Generator):
         details.
         """
         result = []
-        try:
-            relations = obj.getRelationSet()
-        except:
-            relations = []
-        for relation in relations:
-            if relation.getRelationType() ==  pyatspi.RELATION_LABELLED_BY:
-                labelledBy = relation.getTarget(0)
-                result.extend(self.generate(labelledBy, **args))
-                break
+        labels = self._script.utilities.labelsForObject(obj)
+        for label in labels:
+            result.extend(self.generate(label, **args))
+            break
+
         if not result:
             # NOTE: there is no REAL_ROLE_SCROLL_PANE in formatting.py
             # because currently fallback to the default formatting.
