@@ -38,6 +38,7 @@ from . import keybindings
 from . import keynames
 from . import messages
 from . import orca_state
+from . import script_manager
 from . import settings
 
 KEYBOARD_EVENT     = "keyboard"
@@ -90,6 +91,119 @@ class KeyboardEvent(InputEvent):
     TYPE_PUNCTUATION      = "punctuation"
     TYPE_SPACE            = "space"
 
+    GDK_PUNCTUATION_KEYS = [Gdk.KEY_acute,
+                            Gdk.KEY_ampersand,
+                            Gdk.KEY_apostrophe,
+                            Gdk.KEY_asciicircum,
+                            Gdk.KEY_asciitilde,
+                            Gdk.KEY_asterisk,
+                            Gdk.KEY_at,
+                            Gdk.KEY_backslash,
+                            Gdk.KEY_bar,
+                            Gdk.KEY_braceleft,
+                            Gdk.KEY_braceright,
+                            Gdk.KEY_bracketleft,
+                            Gdk.KEY_bracketright,
+                            Gdk.KEY_brokenbar,
+                            Gdk.KEY_cedilla,
+                            Gdk.KEY_cent,
+                            Gdk.KEY_colon,
+                            Gdk.KEY_comma,
+                            Gdk.KEY_copyright,
+                            Gdk.KEY_currency,
+                            Gdk.KEY_degree,
+                            Gdk.KEY_diaeresis,
+                            Gdk.KEY_dollar,
+                            Gdk.KEY_EuroSign,
+                            Gdk.KEY_equal,
+                            Gdk.KEY_exclam,
+                            Gdk.KEY_exclamdown,
+                            Gdk.KEY_grave,
+                            Gdk.KEY_greater,
+                            Gdk.KEY_guillemotleft,
+                            Gdk.KEY_guillemotright,
+                            Gdk.KEY_hyphen,
+                            Gdk.KEY_less,
+                            Gdk.KEY_macron,
+                            Gdk.KEY_minus,
+                            Gdk.KEY_notsign,
+                            Gdk.KEY_numbersign,
+                            Gdk.KEY_paragraph,
+                            Gdk.KEY_parenleft,
+                            Gdk.KEY_parenright,
+                            Gdk.KEY_percent,
+                            Gdk.KEY_period,
+                            Gdk.KEY_periodcentered,
+                            Gdk.KEY_plus,
+                            Gdk.KEY_plusminus,
+                            Gdk.KEY_question,
+                            Gdk.KEY_questiondown,
+                            Gdk.KEY_quotedbl,
+                            Gdk.KEY_quoteleft,
+                            Gdk.KEY_quoteright,
+                            Gdk.KEY_registered,
+                            Gdk.KEY_section,
+                            Gdk.KEY_semicolon,
+                            Gdk.KEY_slash,
+                            Gdk.KEY_sterling,
+                            Gdk.KEY_underscore,
+                            Gdk.KEY_yen]
+
+    GDK_ACCENTED_LETTER_KEYS = [Gdk.KEY_Aacute,
+                                Gdk.KEY_aacute,
+                                Gdk.KEY_Acircumflex,
+                                Gdk.KEY_acircumflex,
+                                Gdk.KEY_Adiaeresis,
+                                Gdk.KEY_adiaeresis,
+                                Gdk.KEY_Agrave,
+                                Gdk.KEY_agrave,
+                                Gdk.KEY_Aring,
+                                Gdk.KEY_aring,
+                                Gdk.KEY_Atilde,
+                                Gdk.KEY_atilde,
+                                Gdk.KEY_Ccedilla,
+                                Gdk.KEY_ccedilla,
+                                Gdk.KEY_Eacute,
+                                Gdk.KEY_eacute,
+                                Gdk.KEY_Ecircumflex,
+                                Gdk.KEY_ecircumflex,
+                                Gdk.KEY_Ediaeresis,
+                                Gdk.KEY_ediaeresis,
+                                Gdk.KEY_Egrave,
+                                Gdk.KEY_egrave,
+                                Gdk.KEY_Iacute,
+                                Gdk.KEY_iacute,
+                                Gdk.KEY_Icircumflex,
+                                Gdk.KEY_icircumflex,
+                                Gdk.KEY_Idiaeresis,
+                                Gdk.KEY_idiaeresis,
+                                Gdk.KEY_Igrave,
+                                Gdk.KEY_igrave,
+                                Gdk.KEY_Ntilde,
+                                Gdk.KEY_ntilde,
+                                Gdk.KEY_Oacute,
+                                Gdk.KEY_oacute,
+                                Gdk.KEY_Ocircumflex,
+                                Gdk.KEY_ocircumflex,
+                                Gdk.KEY_Odiaeresis,
+                                Gdk.KEY_odiaeresis,
+                                Gdk.KEY_Ograve,
+                                Gdk.KEY_ograve,
+                                Gdk.KEY_Ooblique,
+                                Gdk.KEY_ooblique,
+                                Gdk.KEY_Otilde,
+                                Gdk.KEY_otilde,
+                                Gdk.KEY_Uacute,
+                                Gdk.KEY_uacute,
+                                Gdk.KEY_Ucircumflex,
+                                Gdk.KEY_ucircumflex,
+                                Gdk.KEY_Udiaeresis,
+                                Gdk.KEY_udiaeresis,
+                                Gdk.KEY_Ugrave,
+                                Gdk.KEY_ugrave,
+                                Gdk.KEY_Yacute,
+                                Gdk.KEY_yacute]
+
     def __init__(self, event):
         """Creates a new InputEvent of type KEYBOARD_EVENT.
 
@@ -101,10 +215,9 @@ class KeyboardEvent(InputEvent):
         self.id = event.id
         self.type = event.type
         self.hw_code = event.hw_code
-        self.modifiers = event.modifiers
+        self.modifiers = event.modifiers & Gdk.ModifierType.MODIFIER_MASK
         self.event_string = event.event_string
-        self.keyval_name = ""
-        self.is_text = event.is_text
+        self.keyval_name = Gdk.keyval_name(event.id)
         self.timestamp = event.timestamp
         self.is_duplicate = self in [orca_state.lastInputEvent,
                                      orca_state.lastNonModifierKeyEvent]
@@ -120,25 +233,30 @@ class KeyboardEvent(InputEvent):
         self._result_reason = None
         self._bypassOrca = None
 
+        # Some implementors don't populate this field at all. More often than not,
+        # the event_string and the keyval_name coincide for input events.
+        if not self.event_string:
+            self.event_string = self.keyval_name
+
+        # Some implementors do populate the field, but with the keyname rather than
+        # the printable character. This messes us up with punctuation and other symbols.
+        if len(self.event_string) > 1 \
+           and (self.id in KeyboardEvent.GDK_PUNCTUATION_KEYS or \
+                self.id in KeyboardEvent.GDK_ACCENTED_LETTER_KEYS):
+            self.event_string = chr(self.id)
+
         if self._script:
-            self._script.checkKeyboardEventData(self)
             self._app = self._script.app
             if not self._window:
-                self._window = self._script.utilities.activeWindow()
+                self._window = orca_state.activeWindow = self._script.utilities.activeWindow()
+                msg = 'INPUT EVENT: Updated window and active window to %s' % self._window
+                debug.println(debug.LEVEL_INFO, msg, True)
 
-        # Control characters come through as control characters, so we
-        # just turn them into their ASCII equivalent.  NOTE that the
-        # upper case ASCII characters will be used (e.g., ctrl+a will
-        # be turned into the string "A").  All these checks here are
-        # to just do some sanity checking before doing the
-        # conversion. [[[WDW - this is making assumptions about
-        # mapping ASCII control characters to UTF-8.]]]
-        #
-        if (self.modifiers & keybindings.CTRL_MODIFIER_MASK) \
-            and (not self.is_text) and (len(self.event_string) == 1):
-            value = ord(self.event_string[0])
-            if value < 32:
-                self.event_string = chr(value + 0x40)
+        if self._window and self._app != self._window.getApplication():
+            self._script = script_manager.getManager().getScript(self._window.getApplication())
+            self._app = self._script.app
+            msg = 'INPUT EVENT: Updated script to %s' % self._script
+            debug.println(debug.LEVEL_INFO, msg, True)
 
         if self.is_duplicate:
             KeyboardEvent.duplicateCount += 1
@@ -270,19 +388,42 @@ class KeyboardEvent(InputEvent):
         return False
 
     def __str__(self):
-        return ("KEYBOARD_EVENT:  type=%d\n" % self.type) \
-             + ("                 id=%d\n" % self.id) \
-             + ("                 hw_code=%d\n" % self.hw_code) \
-             + ("                 modifiers=%d\n" % self.modifiers) \
-             + ("                 event_string=(%s)\n" % self.event_string) \
-             + ("                 keyval_name=(%s)\n" % self.keyval_name) \
-             + ("                 is_text=%s\n" % self.is_text) \
+        if self._shouldObscure():
+            keyid = hw_code = modifiers = event_string = keyval_name = key_type = "*"
+        else:
+            keyid = self.id
+            hw_code = self.hw_code
+            modifiers = self.modifiers
+            event_string = self.event_string
+            keyval_name = self.keyval_name
+            key_type = self.keyType
+
+        return ("KEYBOARD_EVENT:  type=%s\n" % self.type.value_name.upper()) \
+             + ("                 id=%s\n" % keyid) \
+             + ("                 hw_code=%s\n" % hw_code) \
+             + ("                 modifiers=%s\n" % modifiers) \
+             + ("                 event_string=(%s)\n" % event_string) \
+             + ("                 keyval_name=(%s)\n" % keyval_name) \
              + ("                 timestamp=%d\n" % self.timestamp) \
              + ("                 time=%f\n" % time.time()) \
-             + ("                 keyType=%s\n" % self.keyType) \
+             + ("                 keyType=%s\n" % key_type) \
              + ("                 clickCount=%s\n" % self._clickCount) \
              + ("                 shouldEcho=%s\n" % self.shouldEcho)
 
+    def _shouldObscure(self):
+        if not (self._obj and self._obj.getRole() == pyatspi.ROLE_PASSWORD_TEXT):
+            return False
+
+        if not self.isPrintableKey():
+            return False
+
+        if self.modifiers & keybindings.CTRL_MODIFIER_MASK \
+           or self.modifiers & keybindings.ALT_MODIFIER_MASK \
+           or self.modifiers & keybindings.ORCA_MODIFIER_MASK:
+            return False
+
+        return True
+
     def _isReleaseForLastNonModifierKeyEvent(self):
         last = orca_state.lastNonModifierKeyEvent
         if not last:
@@ -296,6 +437,23 @@ class KeyboardEvent(InputEvent):
 
         return False
 
+    def isReleaseFor(self, other):
+        """Return True if this is the release event for other."""
+
+        if not other:
+            return False
+
+        if not other.isPressedKey() or self.isPressedKey():
+            return False
+
+        return self.id == other.id \
+            and self.hw_code == other.hw_code \
+            and self.modifiers == other.modifiers \
+            and self.event_string == other.event_string \
+            and self.keyval_name == other.keyval_name \
+            and self.keyType == other.keyType \
+            and self._clickCount == other._clickCount
+
     def isNavigationKey(self):
         """Return True if this is a navigation key."""
 
@@ -350,7 +508,7 @@ class KeyboardEvent(InputEvent):
         if self.keyType:
             return self.keyType in KeyboardEvent.TYPE_LOCKING
 
-        lockingKeys = ["Caps_Lock", "Num_Lock", "Scroll_Lock"]
+        lockingKeys = ["Caps_Lock", "Shift_Lock", "Num_Lock", "Scroll_Lock"]
         if not self.event_string in lockingKeys:
             return False
 
@@ -446,6 +604,11 @@ class KeyboardEvent(InputEvent):
 
         return self.event_string in ["space", " "]
 
+    def isFromApplication(self, app):
+        """Return True if this is associated with the specified app."""
+
+        return self._app == app
+
     def isCharacterEchoable(self):
         """Returns True if the script will echo this event as part of
         character echo. We do this to not double-echo a given printable
@@ -470,6 +633,8 @@ class KeyboardEvent(InputEvent):
 
         if self.event_string == "Caps_Lock":
             mod = pyatspi.MODIFIER_SHIFTLOCK
+        elif self.event_string == "Shift_Lock":
+            mod = pyatspi.MODIFIER_SHIFT
         elif self.event_string == "Num_Lock":
             mod = pyatspi.MODIFIER_NUMLOCK
         else:
@@ -611,7 +776,11 @@ class KeyboardEvent(InputEvent):
         """Processes this input event."""
 
         startTime = time.time()
-        data = "'%s' (%d)" % (self.event_string, self.hw_code)
+        if not self._shouldObscure():
+            data = "'%s' (%d)" % (self.event_string, self.hw_code)
+        else:
+            data = "(obscured)"
+
         if self.is_duplicate:
             data = '%s DUPLICATE EVENT #%i' % (data, KeyboardEvent.duplicateCount)
 
@@ -648,7 +817,8 @@ class KeyboardEvent(InputEvent):
         """Processes this input event."""
 
         if self._bypassOrca:
-            if self.event_string == "Caps_Lock" \
+            if (self.event_string == "Caps_Lock" \
+                or self.event_string == "Shift_Lock") \
                and self.type == pyatspi.KEY_PRESSED_EVENT:
                     self._lock_mod()
                     self.keyType = KeyboardEvent.TYPE_LOCKING
@@ -694,10 +864,9 @@ class KeyboardEvent(InputEvent):
         return False, 'Unaddressed case'
 
     def _lock_mod(self):
-        def lock_mod(modifiers):
+        def lock_mod(modifiers, modifier):
             def lockit():
                 try:
-                    modifier = (1 << pyatspi.MODIFIER_SHIFTLOCK)
                     if modifiers & modifier:
                         lock = pyatspi.KEY_UNLOCKMODIFIERS
                         debug.println(debug.LEVEL_INFO, "Locking capslock", True)
@@ -711,8 +880,16 @@ class KeyboardEvent(InputEvent):
                         "at-spi2-core >= 2.32 is needed for triggering capslock", True)
                     pass
             return lockit
+        if self.event_string == "Caps_Lock":
+            modifier = 1 << pyatspi.MODIFIER_SHIFTLOCK
+        elif self.event_string == "Shift_Lock":
+            modifier = 1 << pyatspi.MODIFIER_SHIFT
+        else:
+            msg = "Unknown locking key %s" % self.event_string
+            debug.println(debug.LEVEL_WARNING, msg, True)
+            return
         debug.println(debug.LEVEL_INFO, "Scheduling capslock", True)
-        GLib.timeout_add(1, lock_mod(self.modifiers))
+        GLib.timeout_add(1, lock_mod(self.modifiers, modifier))
 
     def _consume(self):
         startTime = time.time()
@@ -769,7 +946,7 @@ class MouseButtonEvent(InputEvent):
         self.pressed = event.type.endswith('p')
         self.button = event.type[len("mouse:button:"):-1]
         self._script = orca_state.activeScript
-        self.window = None
+        self.window = orca_state.activeWindow
         self.obj = None
 
         if self.pressed:
@@ -778,7 +955,9 @@ class MouseButtonEvent(InputEvent):
         if not self._script:
             return
 
-        self.window = self._script.utilities.activeWindow()
+        if not self._script.utilities.canBeActiveWindow(self.window):
+            self.window = self._script.utilities.activeWindow()
+
         if not self.window:
             return
 
