@@ -130,7 +130,7 @@ class Script(Gecko.Script):
 
         super().locusOfFocusChanged(event, oldFocus, newFocus)
 
-    def useFocusMode(self, obj):
+    def useFocusMode(self, obj, prevObj=None):
         if self.utilities.isEditableMessage(obj):
             msg = "THUNDERBIRD: Using focus mode for editable message %s" % obj
             debug.println(debug.LEVEL_INFO, msg, True)
@@ -173,8 +173,6 @@ class Script(Gecko.Script):
             return
 
         self._lastAutoComplete = ""
-        self.pointOfReference['lastAutoComplete'] = None
-
         obj = event.source
         if self.spellcheck.isAutoFocusEvent(event):
             orca.setLocusOfFocus(event, event.source, False)
@@ -196,6 +194,9 @@ class Script(Gecko.Script):
         if self.utilities.isEditableMessage(event.source):
             return
 
+        if self.inFocusMode():
+            return
+
         obj = event.source
         if self.utilities.isDocument(obj) and not event.detail1:
             try:
@@ -321,7 +322,6 @@ class Script(Gecko.Script):
             if hasSelection or isSystemEvent:
                 speech.speak(event.any_data)
                 self._lastAutoComplete = event.any_data
-                self.pointOfReference['lastAutoComplete'] = hash(obj)
                 return
 
         super().onTextInserted(event)
