@@ -509,6 +509,12 @@ class Script(default.Script):
 
         return super().skipObjectEvent(event)
 
+    def presentationInterrupt(self):
+        super().presentationInterrupt()
+        msg = "WEB: Flushing live region messages"
+        debug.println(debug.LEVEL_INFO, msg, True)
+        self.liveRegionManager.flushMessages()
+
     def consumesKeyboardEvent(self, keyboardEvent):
         """Returns True if the script will consume this keyboard event."""
 
@@ -772,7 +778,7 @@ class Script(default.Script):
 
         return self._browseModeIsSticky
 
-    def useFocusMode(self, obj):
+    def useFocusMode(self, obj, prevObj=None):
         """Returns True if we should use focus mode in obj."""
 
         if self._focusModeIsSticky:
@@ -791,6 +797,9 @@ class Script(default.Script):
             debug.println(debug.LEVEL_INFO, msg, True)
             return False
 
+        if prevObj and self.utilities.isDead(prevObj):
+            prevObj = None
+
         if not _settingsManager.getSetting('caretNavTriggersFocusMode') \
            and self._lastCommandWasCaretNav:
             msg = "WEB: Not using focus mode due to caret nav settings"
@@ -879,7 +888,7 @@ class Script(default.Script):
             priorObj, priorOffset = self.utilities.getPriorContext()
 
         obj, offset = self.utilities.getCaretContext(documentFrame=None)
-        contents = self.utilities.getLineContentsAtOffset(obj, offset)
+        contents = self.utilities.getLineContentsAtOffset(obj, offset, useCache=not isEditable)
         self.speakContents(contents, priorObj=priorObj)
 
     def presentObject(self, obj, **args):
@@ -890,15 +899,20 @@ class Script(default.Script):
         priorObj = args.get("priorObj")
         if self._lastCommandWasCaretNav or args.get("includeContext"):
             priorObj, priorOffset = self.utilities.getPriorContext()
+            args["priorObj"] = priorObj
 
         if obj.getRole() == pyatspi.ROLE_ENTRY:
-            utterances = self.speechGenerator.generateSpeech(obj, priorObj=priorObj)
+            utterances = self.speechGenerator.generateSpeech(obj, **args)
             speech.speak(utterances)
             self.updateBraille(obj)
             return
 
+        # We shouldn't use cache in this method, because if the last thing we presented
+        # included this object and offset (e.g. a Say All or Mouse Review), we're in
+        # danger of presented irrelevant context.
+        useCache = False
         offset = args.get("offset", 0)
-        contents = self.utilities.getObjectContentsAtOffset(obj, offset)
+        contents = self.utilities.getObjectContentsAtOffset(obj, offset, useCache)
         self.displayContents(contents)
         self.speakContents(contents, **args)
  
@@ -1052,7 +1066,7 @@ class Script(default.Script):
         """To-be-removed. Returns the string, caretOffset, startOffset."""
 
         if self._inFocusMode or not self.utilities.inDocumentContent(obj) \
-           or obj.getState().contains(pyatspi.STATE_EDITABLE):
+           or self.utilities.isFocusModeWidget(obj):
             return super().getTextLineAtCaret(obj, offset, startOffset, endOffset)
 
         text = self.utilities.queryNonEmptyText(obj)
@@ -1231,6 +1245,17 @@ class Script(default.Script):
             debug.println(debug.LEVEL_INFO, msg, True)
             contents = self.utilities.getObjectContentsAtOffset(newFocus, 0)
             utterances = self.speechGenerator.generateContents(contents)
+        elif self.utilities.caretMovedToSamePageFragment(event, oldFocus):
+            msg = "WEB: Event source %s is same page fragment. Generating line contents." % event.source
+            debug.println(debug.LEVEL_INFO, msg, True)
+            contents = self.utilities.getLineContentsAtOffset(newFocus, 0)
+            utterances = self.speechGenerator.generateContents(contents)
+        elif self._lastCommandWasMouseButton and event \
+             and event.type.startswith("object:text-caret-moved"):
+            msg = "WEB: Last input event was mouse button. Generating line contents."
+            debug.println(debug.LEVEL_INFO, msg, True)
+            contents = self.utilities.getLineContentsAtOffset(newFocus, caretOffset)
+            utterances = self.speechGenerator.generateContents(contents)
         else:
             msg = "WEB: New focus %s is not a special case. Generating speech." % newFocus
             debug.println(debug.LEVEL_INFO, msg, True)
@@ -1313,7 +1338,20 @@ class Script(default.Script):
         if not obj or self.utilities.isZombie(obj):
             self.utilities.clearCaretContext()
 
-        shouldPresent = self.utilities.isShowingOrVisible(event.source)
+        shouldPresent = True
+        if not self.utilities.isShowingOrVisible(event.source):
+            shouldPresent = False
+            msg = "WEB: Not presenting because source is not showing or visible"
+            debug.println(debug.LEVEL_INFO, msg, True)
+        elif not self.utilities.documentFrameURI(event.source):
+            shouldPresent = False
+            msg = "WEB: Not presenting because source lacks URI"
+            debug.println(debug.LEVEL_INFO, msg, True)
+        elif not event.detail1 and self._inFocusMode and not self.utilities.isZombie(obj):
+            shouldPresent = False
+            msg = "WEB: Not presenting due to focus mode for %s" % obj
+            debug.println(debug.LEVEL_INFO, msg, True)
+
         if not _settingsManager.getSetting('onlySpeakDisplayedText') and shouldPresent:
             if event.detail1:
                 self.presentMessage(messages.PAGE_LOADING_START)
@@ -1440,6 +1478,11 @@ class Script(default.Script):
             return True
 
         if self._lastCommandWasMouseButton:
+            if (event.source, event.detail1) == self.utilities.getCaretContext():
+                msg = "WEB: Event is for current caret context."
+                debug.println(debug.LEVEL_INFO, msg, True)
+                return True
+
             msg = "WEB: Event handled: Last command was mouse button"
             debug.println(debug.LEVEL_INFO, msg, True)
             self.utilities.setCaretContext(event.source, event.detail1)
@@ -1466,11 +1509,6 @@ class Script(default.Script):
             debug.println(debug.LEVEL_INFO, msg, True)
             return True
 
-        if self.utilities.textEventIsForNonNavigableTextObject(event):
-            msg = "WEB: Event ignored: Event source is non-navigable text object"
-            debug.println(debug.LEVEL_INFO, msg, True)
-            return True
-
         if self.utilities.textEventIsDueToInsertion(event):
             msg = "WEB: Event handled: Updating position due to insertion"
             debug.println(debug.LEVEL_INFO, msg, True)
@@ -1486,6 +1524,13 @@ class Script(default.Script):
             orca.setLocusOfFocus(event, obj)
             return True
 
+        # We want to do this check after the same-page-fragment check because some
+        # fragments start with non-navigable text objects.
+        if self.utilities.textEventIsForNonNavigableTextObject(event):
+            msg = "WEB: Event ignored: Event source is non-navigable text object"
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return True
+
         if self.utilities.lastInputEventWasPageNav() \
            and not self.utilities.isLink(event.source) \
            and not event.source.getRole() == pyatspi.ROLE_COMBO_BOX:
@@ -1553,12 +1598,11 @@ class Script(default.Script):
 
         if not _settingsManager.getSetting('caretNavigationEnabled') \
            or self._inFocusMode or isEditable:
+            msg = "WEB: Setting locusOfFocus, context to: %s, %i" % (event.source, event.detail1)
+            debug.println(debug.LEVEL_INFO, msg, True)
             self.utilities.setCaretContext(event.source, event.detail1)
             notify = event.source.getState().contains(pyatspi.STATE_FOCUSED)
             orca.setLocusOfFocus(event, event.source, notify)
-            msg = "WEB: Setting locusOfFocus, context to: %s, %i" % \
-                  (event.source, event.detail1)
-            debug.println(debug.LEVEL_INFO, msg, True)
             return False
 
         self.utilities.setCaretContext(obj, offset)
@@ -1647,20 +1691,16 @@ class Script(default.Script):
             debug.println(debug.LEVEL_INFO, msg, True)
             return True
 
+        if self.utilities.handleEventFromContextReplicant(event, event.any_data):
+            msg = "WEB: Event handled by updating locusOfFocus and context to child."
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return True
+
         obj, offset = self.utilities.getCaretContext(getZombieReplicant=False)
         msg = "WEB: Context: %s, %i (focus: %s)" % (obj, offset, orca_state.locusOfFocus)
         debug.println(debug.LEVEL_INFO, msg, True)
 
         if self.utilities.isZombie(obj):
-            if self.utilities.isSameObject(obj, event.any_data, comparePaths=True, ignoreNames=True):
-                path, role, name = self.utilities.getCaretContextPathRoleAndName()
-                notify = event.any_data.name != name
-                msg = "WEB: Event handled by updating locusOfFocus and context"
-                debug.println(debug.LEVEL_INFO, msg, True)
-                orca.setLocusOfFocus(event, event.any_data, notify)
-                self.utilities.setCaretContext(event.any_data, offset)
-                return True
-
             obj, offset = self.utilities.getCaretContext(getZombieReplicant=True)
             if not obj:
                 if self._inFocusMode:
@@ -1707,12 +1747,6 @@ class Script(default.Script):
                 self.utilities.setCaretContext(focused, 0)
             return True
 
-        if childRole == pyatspi.ROLE_DIALOG:
-            msg = "WEB: Setting locusOfFocus to event.any_data"
-            debug.println(debug.LEVEL_INFO, msg, True)
-            orca.setLocusOfFocus(event, event.any_data)
-            return True
-
         if self.lastMouseRoutingTime and 0 < time.time() - self.lastMouseRoutingTime < 1:
             utterances = []
             utterances.append(messages.NEW_ITEM_ADDED)
@@ -1727,6 +1761,11 @@ class Script(default.Script):
     def onDocumentLoadComplete(self, event):
         """Callback for document:load-complete accessibility events."""
 
+        if self.utilities.getDocumentForObject(event.source.parent):
+            msg = "WEB: Ignoring: Event source is nested document"
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return True
+
         msg = "WEB: Updating loading state and resetting live regions"
         debug.println(debug.LEVEL_INFO, msg, True)
         self._loadingDocumentContent = False
@@ -1736,6 +1775,11 @@ class Script(default.Script):
     def onDocumentLoadStopped(self, event):
         """Callback for document:load-stopped accessibility events."""
 
+        if self.utilities.getDocumentForObject(event.source.parent):
+            msg = "WEB: Ignoring: Event source is nested document"
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return True
+
         msg = "WEB: Updating loading state"
         debug.println(debug.LEVEL_INFO, msg, True)
         self._loadingDocumentContent = False
@@ -1744,6 +1788,11 @@ class Script(default.Script):
     def onDocumentReload(self, event):
         """Callback for document:reload accessibility events."""
 
+        if self.utilities.getDocumentForObject(event.source.parent):
+            msg = "WEB: Ignoring: Event source is nested document"
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return True
+
         msg = "WEB: Updating loading state"
         debug.println(debug.LEVEL_INFO, msg, True)
         self._loadingDocumentContent = True
@@ -1764,6 +1813,17 @@ class Script(default.Script):
 
         return False
 
+    def onFocus(self, event):
+        """Callback for focus: accessibility events."""
+
+        # We should get proper state-changed events for these.
+        if self.utilities.inDocumentContent(event.source):
+            msg = "WEB: Ignoring because object:state-changed-focused expected."
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return True
+
+        return False
+
     def onFocusedChanged(self, event):
         """Callback for object:state-changed:focused accessibility events."""
 
@@ -1806,6 +1866,11 @@ class Script(default.Script):
             orca.setLocusOfFocus(event, event.source)
             return True
 
+        if self.utilities.handleEventFromContextReplicant(event, event.source):
+            msg = "WEB: Event handled by updating locusOfFocus and context to source."
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return True
+
         obj, offset = self.utilities.getCaretContext()
         msg = "WEB: Caret context is %s, %i (focus: %s)" \
               % (obj, offset, orca_state.locusOfFocus)
@@ -1970,6 +2035,12 @@ class Script(default.Script):
     def onShowingChanged(self, event):
         """Callback for object:state-changed:showing accessibility events."""
 
+        if event.detail1 and self.utilities.isTopLevelBrowserUIAlert(event.source):
+            msg = "WEB: Event handled: Presenting event source"
+            debug.println(debug.LEVEL_INFO, msg, True)
+            self.presentObject(event.source)
+            return True
+
         if not self.utilities.inDocumentContent(event.source):
             msg = "WEB: Event source is not in document content"
             debug.println(debug.LEVEL_INFO, msg, True)
@@ -2005,6 +2076,11 @@ class Script(default.Script):
             debug.println(debug.LEVEL_INFO, msg, True)
             return True
 
+        if self.utilities.textEventIsDueToDeletion(event):
+            msg = "WEB: Event believed to be due to editable text deletion"
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return False
+
         if self.utilities.textEventIsDueToInsertion(event):
             msg = "WEB: Ignoring event believed to be due to text insertion"
             debug.println(debug.LEVEL_INFO, msg, True)
@@ -2083,8 +2159,6 @@ class Script(default.Script):
             debug.println(debug.LEVEL_INFO, msg, True)
             return True
 
-        # TODO - JD: As an experiment, we're stopping these at the event manager.
-        # If that works, this can be removed.
         if self.utilities.eventIsEOCAdded(event):
             msg = "WEB: Ignoring: Event was for embedded object char"
             debug.println(debug.LEVEL_INFO, msg, True)
